@@ -18,26 +18,33 @@
  * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
  *
  * $LastChangedBy: bastian $
- * $LastChangedDate: 2007-05-13 16:29:34 +0000 (So, 13 Mai 2007) $
- * $LastChangedRevision: 1434 $
+ * $LastChangedDate: 2008-06-24 12:22:06 +0200 (Di, 24 Jun 2008) $
+ * $LastChangedRevision: 1668 $
  */
 
+#define _GNU_SOURCE
+
 #include <config.h>
 
 #include "download.h"
 #include "execute.h"
 #include "frontend.h"
+#include "gpg.h"
 #include "install.h"
 #include "message.h"
 #include "log.h"
 #include "suite.h"
+#include "target.h"
 
 #include <debian-installer.h>
 
+#include <ctype.h>
+#include <errno.h>
 #include <getopt.h>
 #include <libgen.h>
 #include <limits.h>
 #include <sys/stat.h>
+#include <sys/statvfs.h>
 #include <sys/types.h>
 #include <stdarg.h>
 #include <stdbool.h>
@@ -46,7 +53,18 @@
 #include <string.h>
 #include <unistd.h>
 
-const char *mirror;
+const char *target_root;
+FILE *foreign_script;
+
+static enum
+{
+    MIRROR_SCHEME_HTTP,
+    MIRROR_SCHEME_FTP,
+    MIRROR_SCHEME_FILE,
+    MIRROR_SCHEME_SSH,
+} mirror_scheme;
+static char mirror_host[128];
+static char mirror_path[256];
 
 #ifdef DEB_ARCH
 const char default_arch[] = DEB_ARCH;
@@ -61,7 +79,9 @@ enum
   GETOPT_ALLOW_UNAUTHENTICATED,
   GETOPT_DEBUG,
   GETOPT_EXCLUDE,
+  GETOPT_FOREIGN,
   GETOPT_INCLUDE,
+  GETOPT_SUITE_CONFIG,
   GETOPT_VARIANT,
   GETOPT_VERSION,
 };
@@ -74,12 +94,13 @@ static struct option const long_opts[] =
   {"debug", no_argument, 0, GETOPT_DEBUG},
   {"download-only", no_argument, 0, 'd'},
   {"exclude", required_argument, 0, GETOPT_EXCLUDE},
+  {"foreign", no_argument, 0, GETOPT_FOREIGN},
   {"flavour", required_argument, 0, 'f'},
   {"helperdir", required_argument, 0, 'H'},
   {"include", required_argument, 0, GETOPT_INCLUDE},
   {"keyring", required_argument, 0, 'k'},
   {"quiet", no_argument, 0, 'q'},
-  {"suite-config", required_argument, 0, 's'},
+  {"suite-config", required_argument, 0, GETOPT_SUITE_CONFIG},
   {"variant", required_argument, 0, GETOPT_VARIANT},
   {"verbose", no_argument, 0, 'v'},
   {"help", no_argument, 0, 'h'},
@@ -94,39 +115,197 @@ int frontend_download (const char *source, const char *target)
   char buf[1024];
   int ret;
 
-  if (!strncmp (mirror, "http://", 7) || !strncmp (mirror, "ftp://", 6))
-    snprintf (buf, 1024, "wget -q -O %s %s/%s", target, mirror, source);
-  else if (!strncmp (mirror, "file://", 7))
-    snprintf (buf, 1024, "cp %s/%s %s", mirror + sizeof ("file://") - 1, source, target);
+  switch (mirror_scheme)
+  {
+    case MIRROR_SCHEME_HTTP:
+    case MIRROR_SCHEME_FTP:
+      {
+        const char *scheme = "http";
+
+        if (mirror_scheme == MIRROR_SCHEME_FTP)
+          scheme = "ftp";
+
+        snprintf (buf, sizeof buf, "wget -q -O %s %s://%s%s/%s", target, scheme, mirror_host, mirror_path, source);
+      }
+      break;
+
+    case MIRROR_SCHEME_FILE:
+      snprintf (buf, sizeof buf, "cp %s/%s %s", mirror_path, source, target);
+      break;
+
+    case MIRROR_SCHEME_SSH:
+      snprintf (buf, sizeof buf, "ssh -o BatchMode=yes %s 'cat %s/%s' > %s", mirror_host, mirror_path, source, target);
+  }
+
   ret = execute (buf);
 
   return WEXITSTATUS (ret);
 }
 
-static int check_mirror (void)
+static void check_mirror_scheme (const char **input)
 {
-  if (!strncmp (mirror, "http://", 7))
-    return 0;
-  else if (!strncmp (mirror, "ftp://", 6))
-    return 0;
-  else if (!strncmp (mirror, "file://", 7))
-    return 0;
-  return 1;
+  const char *temp = *input;
+
+  while (*temp >= 'a' && *temp <= 'z')
+    temp++;
+
+  if (*temp++ != ':' || *temp++ != '/' || *temp++ != '/')
+    log_text (DI_LOG_LEVEL_ERROR, "Invalid mirror: can't find scheme");
+
+  if (!strncmp (*input, "http", 4))
+    mirror_scheme = MIRROR_SCHEME_HTTP;
+  else if (!strncmp (*input, "ftp", 3))
+    mirror_scheme = MIRROR_SCHEME_FTP;
+  else if (!strncmp (*input, "file", 4))
+    mirror_scheme = MIRROR_SCHEME_FILE;
+  else if (!strncmp (*input, "ssh", 3))
+    mirror_scheme = MIRROR_SCHEME_SSH;
+  else
+    log_text (DI_LOG_LEVEL_ERROR, "Invalid mirror: scheme not supported");
+
+  *input = temp;
 }
 
-static int finish_etc (void)
+static void check_mirror_host (const char **input)
+{
+  const char *temp = *input;
+
+  switch (mirror_scheme)
+  {
+    case MIRROR_SCHEME_FILE:
+      if (*temp != '/')
+        log_text (DI_LOG_LEVEL_ERROR, "Invalid mirror: file scheme must not include a host");
+      break;
+
+    default:
+      if (*temp == '/')
+        log_text (DI_LOG_LEVEL_ERROR, "Invalid mirror: scheme must include a host");
+      while (*temp && *temp != '/')
+        temp++;
+      strncpy (mirror_host, *input, temp - *input);
+  }
+
+  *input = temp;
+}
+
+static void check_mirror_path (const char *input)
+{
+  size_t len = strlen (input);
+
+  while (len > 0 && input[len - 1] == '/')
+    len--;
+
+  strncpy (mirror_path, input, len);
+}
+
+static void check_mirror (const char *mirror)
+{
+  const char *temp = mirror;
+
+  check_mirror_scheme (&temp);
+  check_mirror_host (&temp);
+  check_mirror_path (temp);
+}
+
+static void check_permission (void)
+{
+  if (getuid ())
+    log_text (DI_LOG_LEVEL_ERROR, "Need root privileges");
+}
+
+static void check_target (const char *target)
 {
-  char buf1[PATH_MAX], buf2[PATH_MAX], buf3[PATH_MAX];
-  FILE *in, *out;
-  size_t len;
   struct stat s;
+  struct statvfs sv;
+
+  if (stat (target, &s) == 0)
+  {
+    if (!(S_ISDIR (s.st_mode)))
+      log_text (DI_LOG_LEVEL_ERROR, "Target exists but is no directory");
+  }
+  else if (errno == ENOENT)
+  {
+    if (mkdir (target, 0777))
+      log_text (DI_LOG_LEVEL_ERROR, "Failed to creae target");
+  }
+  else
+    log_text (DI_LOG_LEVEL_ERROR, "Target check failed: %s", strerror (errno));
+
+  if (statvfs (target, &sv) == 0)
+  {
+    if (sv.f_flag & ST_RDONLY)
+      log_text (DI_LOG_LEVEL_ERROR, "Target is readonly");
+    if (sv.f_flag & ST_NODEV)
+      log_text (DI_LOG_LEVEL_ERROR, "Target disallows device special files");
+    if (sv.f_flag & ST_NOEXEC)
+      log_text (DI_LOG_LEVEL_ERROR, "Target disallows program execution");
+  }
+  else
+    log_text (DI_LOG_LEVEL_ERROR, "Target check failed: %s", strerror (errno));
+
+  target_root = canonicalize_file_name(target);
+  if (!target_root)
+    log_text (DI_LOG_LEVEL_ERROR, "Target check failed: %s", strerror (errno));
+}
+
+static int finish_etc_apt_sources_list (void)
+{
+  char file[PATH_MAX];
+  FILE *out;
+  char line[1024];
+  const char *scheme = NULL;
+
+  switch (mirror_scheme)
+  {
+    case MIRROR_SCHEME_HTTP:
+      scheme = "http";
+      break;
+    case MIRROR_SCHEME_FTP:
+      scheme = "ftp";
+      break;
+    case MIRROR_SCHEME_FILE:
+      return 0;
+    case MIRROR_SCHEME_SSH:
+      scheme = "ssh";
+      break;
+  }
+
+  snprintf (line, sizeof line, "deb %s://%s%s %s main", scheme, mirror_host, mirror_path, suite_name);
+
+  if (foreign_script)
+    fprintf (foreign_script, "\necho '%s' > /etc/apt/sources.list\n\n", line);
+  else
+  {
+    log_text (DI_LOG_LEVEL_MESSAGE, "Writing apt sources.list");
+
+    strcpy (file, target_root);
+    strcat (file, "/etc/apt/sources.list");
+
+    out = fopen (file, "w");
+    if (!out)
+      return 1;
+
+    if (!fprintf (out, "%s\n", line))
+      return 1;
+
+    if (fclose (out))
+      return 1;
+  }
+
+  return 0;
+}
+
+static int finish_etc_hosts (void)
+{
+  char file[PATH_MAX];
+  FILE *out;
 
   log_text (DI_LOG_LEVEL_MESSAGE, "Writing hosts");
 
-  strcpy (buf1, install_root);
-  strcat (buf1, "/etc/hosts");
+  strcpy (file, target_root);
+  strcat (file, "/etc/hosts");
 
-  out = fopen (buf1, "w");
+  out = fopen (file, "w");
   if (!out)
     return 1;
 
@@ -136,49 +315,50 @@ static int finish_etc (void)
   if (fclose (out))
     return 1;
 
-  strcpy (buf2, install_root);
-  strcpy (buf1, "/etc/resolv.conf");
-  strcat (buf2, "/etc/resolv.conf");
+  return 0;
+}
 
-  if (!stat (buf1, &s))
+static int finish_etc_resolv_conf (void)
+{
+  char file_in[PATH_MAX], file_out[PATH_MAX], buf[1024];
+  FILE *in, *out;
+  struct stat s;
+
+  strcpy (file_in, "/etc/resolv.conf");
+  strcpy (file_out, target_root);
+  strcat (file_out, "/etc/resolv.conf");
+
+  if (!stat (file_in, &s))
   {
     log_text (DI_LOG_LEVEL_MESSAGE, "Writing resolv.conf");
 
-    in = fopen (buf1, "r");
-    out = fopen (buf2, "w");
+    in = fopen (file_in, "r");
+    out = fopen (file_out, "w");
     if (!in || !out)
       return 1;
 
     while (1)
     {
-      len = fread (buf3, 1, sizeof (buf3), in);
+      size_t len = fread (buf, 1, sizeof buf, in);
       if (!len)
         break;
-      fwrite (buf3, 1, len, out);
+      fwrite (buf, 1, len, out);
     }
 
     if (fclose (in) || fclose (out))
       return 1;
   }
 
-  log_text (DI_LOG_LEVEL_MESSAGE, "Writing apt sources.list");
-
-  strcpy (buf1, install_root);
-  strcat (buf1, "/etc/apt/sources.list");
-
-  out = fopen (buf1, "w");
-  if (!out)
-    return 1;
-
-  if (!fprintf (out, "deb %s %s main\n", mirror, suite_name))
-    return 1;
-
-  if (fclose (out))
-    return 1;
-
   return 0;
 }
 
+static int finish_etc (void)
+{
+  return finish_etc_apt_sources_list () ||
+         finish_etc_hosts () ||
+         finish_etc_resolv_conf ();
+}
+
 static const char *generate_configdir ()
 {
 #ifdef CONFIGDIR_BINARY
@@ -196,6 +376,97 @@ static const char *generate_configdir ()
 #endif
 }
 
+static int foreign_finish ()
+{
+  char buf1[PATH_MAX], buf2[PATH_MAX];
+  int ret;
+  
+  snprintf (buf1, sizeof buf1, "%s/sbin/init", target_root);
+  snprintf (buf2, sizeof buf2, "%s/sbin/init.foreign", target_root);
+
+  ret = rename (buf1, buf2);
+  if (ret)
+    return ret;
+
+  ret = symlink ("cdebootstrap-foreign", buf1);
+  if (ret)
+    return ret;
+
+  fputs ("\
+rm /sbin/init\n\
+dpkg-divert --remove --local --divert /sbin/init.REAL /sbin/init\n\
+rm /sbin/cdebootstrap-foreign\n\
+", foreign_script);
+
+  snprintf (buf1, sizeof buf1, "%s/dev/console", target_root);
+
+  if (mknod (buf1, S_IFCHR | 0600, makedev (5, 1)) < 0)
+    log_text (DI_LOG_LEVEL_WARNING, "Creation of /dev/console failed");
+
+  log_text (DI_LOG_LEVEL_INFO, "Second stage installer is available as /sbin/cdebootstrap-foreign or /sbin/init");
+
+  return 0;
+}
+
+static install_simulate_handler foreign_handler;
+static int foreign_handler (const char *command)
+{
+  fputs (command, foreign_script);
+  fputs ("\n", foreign_script);
+  return 0;
+}
+
+static int foreign_init ()
+{
+  char buf[PATH_MAX];
+  int ret;
+  
+  target_create_dir ("/sbin");
+
+  snprintf (buf, sizeof buf, "%s/sbin/cdebootstrap-foreign", target_root);
+  foreign_script = fopen (buf, "w");
+
+  if (!foreign_script)
+    return 1;
+
+  ret = chmod (buf, 0755);
+  if (ret)
+    return ret;
+
+  fputs ("\
+#!/bin/sh\n\
+\n\
+set -e\n\
+\n\
+", foreign_script);
+
+  const char *const *temp = execute_environment_target;
+  while (*temp)
+    fprintf (foreign_script, "export %s\n", *temp++);
+
+  fputs ("\
+\n\
+if [ \"$0\" = /sbin/init ]; then\n\
+  trap 'exec env -u DEBIAN_FRONTEND /bin/sh' ERR\n\
+  trap 'exec env -u DEBIAN_FRONTEND -u PATH /sbin/init' EXIT\n\
+  mount -n -o remount,rw rootfs /\n\
+else\n\
+  trap 'umount /proc || :' EXIT\n\
+  mount -n -t proc none /proc\n\
+fi\n\
+\n\
+if [ ! -e /sbin/init.REAL ]; then\n\
+  rm /sbin/init\n\
+  ln -s /bin/sh /sbin/init\n\
+  dpkg-divert --add --local --divert /sbin/init.REAL /sbin/init\n\
+  mv /sbin/init.foreign /sbin/init.REAL\n\
+fi\n\
+\n\
+", foreign_script);
+
+  return 0;
+}
+
 static void usage (int status)
 {
   if (status != 0)
@@ -217,6 +488,7 @@ Mandatory arguments to long options are mandatory for short options too.\n\
   -d, --download-only          Download packages, but don't perform installation.\n\
       --exclude=A,B,C          Drop packages from the installation list\n\
   -f, --flavour=FLAVOUR        Select the flavour to use.\n\
+      --foreign                Use two stage installer.\n\
   -k, --keyring=KEYRING        Use given keyring.\n\
   -H, --helperdir=DIR          Set the helper directory.\n\
       --include=A,B,C          Install extra packages.\n\
@@ -245,7 +517,7 @@ CONFIGDIR
   exit (status);
 }
 
-int frontend_main (int argc, char **argv, char **envp)
+int frontend_main (int argc, char **argv, char **envp __attribute((unused)))
 {
   int c;
   const char
@@ -254,10 +526,11 @@ int frontend_main (int argc, char **argv, char **envp)
     *flavour = default_flavour,
     *keyring = NULL,
     *helperdir,
+    *mirror = NULL,
     *suite = NULL,
     *suite_config = NULL,
     *target = NULL;
-  bool authentication = true, download_only = false;
+  bool authentication = true, download_only = false, foreign = false;
   di_slist include = { NULL, NULL }, exclude = { NULL, NULL };
   const char *keyringdirs[] =
   {
@@ -299,9 +572,6 @@ int frontend_main (int argc, char **argv, char **envp)
       case 'k':
         keyring = optarg;
         break;
-      case 's':
-        suite_config = optarg;
-        break;
       case 'q':
         if (message_level <= MESSAGE_LEVEL_NORMAL)
           message_level = MESSAGE_LEVEL_QUIET;
@@ -323,6 +593,9 @@ int frontend_main (int argc, char **argv, char **envp)
             di_slist_append (&exclude, i);
         }
         break;
+      case GETOPT_FOREIGN:
+        foreign = true;
+        break;
       case GETOPT_INCLUDE:
         {
           char *l = strdup (optarg);
@@ -330,6 +603,9 @@ int frontend_main (int argc, char **argv, char **envp)
             di_slist_append (&include, i);
         }
         break;
+      case GETOPT_SUITE_CONFIG:
+        suite_config = optarg;
+        break;
       case GETOPT_VARIANT:
         if (!strcmp(optarg, "buildd"))
           flavour = "build";
@@ -388,32 +664,46 @@ warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n\
 
   di_init (basename (program_name));
 
-  if (check_mirror ())
-    log_text (DI_LOG_LEVEL_ERROR, "invalid mirror");
+  umask (022);
 
-  if (!download_only && getuid ())
-    log_text (DI_LOG_LEVEL_ERROR, "need root privileges");
+  check_mirror (mirror);
+  if (!download_only)
+    check_permission ();
+  check_target (target);
 
   log_init ();
 
-  if (install_target_check (target))
-    log_text (DI_LOG_LEVEL_ERROR, "target check");
-  if (suite_init (suite, suite_config, arch, flavour, &include, &exclude, configdir, authentication, keyring, keyringdirs))
-    log_text (DI_LOG_LEVEL_ERROR, "suite init");
-  if (execute_init (envp))
-    log_text (DI_LOG_LEVEL_ERROR, "execute init");
+  if (suite_init (suite, suite_config, arch, flavour, &include, &exclude, configdir, authentication))
+    log_text (DI_LOG_LEVEL_ERROR, "Internal error: suite init");
+  if (gpg_init (keyringdirs, keyring))
+    log_text (DI_LOG_LEVEL_ERROR, "Internal error: gpg init");
   if (download_init ())
-    log_text (DI_LOG_LEVEL_ERROR, "download init");
-  if (!download_only && install_init (helperdir))
-    log_text (DI_LOG_LEVEL_ERROR, "install init");
-
-  if (download (&packages, &allocator, &list))
-    log_text (DI_LOG_LEVEL_ERROR, "download");
-  if (!download_only && install (packages, allocator, list))
-    log_text (DI_LOG_LEVEL_ERROR, "install");
+    log_text (DI_LOG_LEVEL_ERROR, "Internal error: download init");
+  if (foreign && foreign_init ())
+    log_text (DI_LOG_LEVEL_ERROR, "Internal error: foreign init");
 
   if (!download_only)
-    finish_etc ();
+  {
+    install_simulate_handler *tmp = NULL;
+    if (foreign)
+      tmp = foreign_handler;
+    if (install_init (helperdir, tmp))
+      log_text (DI_LOG_LEVEL_ERROR, "Internal error: install init");
+  }
+
+  if (download (&packages, &allocator, &list))
+    log_text (DI_LOG_LEVEL_ERROR, "Internal error: download");
+
+  if (download_only)
+    return 0;
+
+  if (install (packages, allocator, list))
+    log_text (DI_LOG_LEVEL_ERROR, "Internal error: install");
+
+  finish_etc ();
+
+  if (foreign && foreign_finish ())
+    log_text (DI_LOG_LEVEL_ERROR, "Internal error: foreign finish");
 
   return 0;
 }
